package org.amse.ak.schemebuilder.io.reader;

import java.io.File;
import java.io.IOException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.amse.ak.schemebuilder.builder.BuildSchemeState;
import org.amse.ak.schemebuilder.builder.ISchemeBuilder;
import org.amse.ak.schemebuilder.builder.SchemeBuilderException;
import org.amse.ak.schemebuilder.io.SchemeIOInterface;
import org.amse.ak.schemebuilder.model.IActionBlock;
import org.amse.ak.schemebuilder.model.IBlock;
import org.amse.ak.schemebuilder.model.IBlockFactory;
import org.amse.ak.schemebuilder.model.IDoBlock;
import org.amse.ak.schemebuilder.model.IEndBlock;
import org.amse.ak.schemebuilder.model.IForBlock;
import org.amse.ak.schemebuilder.model.IIfBlock;
import org.amse.ak.schemebuilder.model.IInvocationBlock;
import org.amse.ak.schemebuilder.model.IScheme;
import org.amse.ak.schemebuilder.model.ISwitchBlock;
import org.amse.ak.schemebuilder.model.ITryBlock;
import org.amse.ak.schemebuilder.model.IWhileBlock;
import org.amse.ak.schemebuilder.model.impl.BlockFactory;
import org.amse.ak.schemebuilder.model.impl.Method;
import org.amse.ak.schemebuilder.model.impl.Scheme;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class SchemeReader implements SchemeIOInterface, ISchemeBuilder {
    
    private final IBlockFactory myBlockFactory = new BlockFactory();
    
    private final File myFile;
    
    private IScheme myScheme;
    
    private IBlock myCurrentBlock;
    
    private Document myDocument;
    
    private BuildSchemeState myState = BuildSchemeState.BEGIN_STATE;
    
    public SchemeReader(File file) {
        myFile = file;
    }

    public IScheme build() throws SchemeBuilderException {
        
        try {
            DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            myDocument = builder.parse(myFile);
        } catch (SAXException e) {
            throw new SchemeBuilderException("Can't parse file.", e);
        } catch (ParserConfigurationException e) {
            throw new SchemeBuilderException("Can't create parser.", e);
        } catch (IOException e) {
            throw new SchemeBuilderException("Can't open file.", e);
        }
        
        Element scheme = myDocument.getDocumentElement();
        NodeList nodes = scheme.getChildNodes();
        Node m = nodes.item(1);
        if (!METHOD.equals(m.getNodeName())) {
            throw new SchemeBuilderException("Irregular XML file. Excepted node "
                    + METHOD + ", but received " + m.getNodeName() + ".");
        }
        Method method = buildMethod(m);
        Node bb = nodes.item(3);
        if (!BEGIN_BLOCK.equals(bb.getNodeName())) {
            throw new SchemeBuilderException("Irregular XML file. Excepted node "
                    + BEGIN_BLOCK + ", but received " + bb.getNodeName() + ".");
        }
        NamedNodeMap attrs = bb.getAttributes();
        myScheme = new Scheme(method, attrs.getNamedItem(TEXT).getNodeValue());
        myCurrentBlock = myScheme.getBeginBlock();
        int n = nodes.getLength();
        for (int i = 5; i < n; i += 2) {
            visitNode(nodes.item(i));
        }
        return myScheme;
    }
    
    private Method buildMethod(Node method) throws SchemeBuilderException {
        NamedNodeMap attrs = method.getAttributes();
        String methodPath = attrs.getNamedItem(PATH).getNodeValue();
        String methodName = attrs.getNamedItem(NAME).getNodeValue();
        return new Method(methodPath, methodName);
    }
    
    private void visitNode(Node node) throws SchemeBuilderException {
        NamedNodeMap attrs = node.getAttributes();
        String blockType = node.getNodeName();
        String blockText = attrs.getNamedItem(TEXT).getNodeValue();
        if (ACTION_BLOCK.equals(blockType)) {
            IActionBlock ab = myBlockFactory.createActionBlock(blockText);
            addBlock(ab);
        } else if (DO_BLOCK.equals(blockType)) {
            IDoBlock db = myBlockFactory.createDoBlock(blockText);
            addBlock(db);
            myState = BuildSchemeState.FIRST_DO_BLOCK;
            NodeList nodes = node.getChildNodes();
            int n = nodes.getLength();
            for (int i = 1; i < n; i += 2) {
                visitNode(nodes.item(i));
            }
            myState = BuildSchemeState.BEGIN_STATE;
            myCurrentBlock = db;
        } else if (END_BLOCK.equals(blockType)) {
            IEndBlock eb = myBlockFactory.createEndBlock(blockText);
            addBlock(eb);
        } else if (FOR_BLOCK.equals(blockType)) {
            IForBlock fb = myBlockFactory.createForBlock(blockText);
            addBlock(fb);
            myState = BuildSchemeState.FIRST_FOR_BLOCK;
            NodeList nodes = node.getChildNodes();
            int n = nodes.getLength();
            for (int i = 1; i < n; i += 2) {
                visitNode(nodes.item(i));
            }
            myState = BuildSchemeState.BEGIN_STATE;
            myCurrentBlock = fb;
        } else if (IF_BLOCK.equals(blockType)) {
            IIfBlock ib = myBlockFactory.createIfBlock(blockText);
            addBlock(ib);
            NodeList nodes = node.getChildNodes();
            Node yes = nodes.item(1);
            if (!YES.equals(yes.getNodeName())) {
                throw new SchemeBuilderException("Irregular XML file. Excepted node "
                        + YES + ", but received " + yes.getNodeName() + ".");
            }
            myState = BuildSchemeState.YES_BLOCK;
            NodeList yesNodes = yes.getChildNodes();
            int n = yesNodes.getLength();
            for (int i = 1; i < n; i += 2) {
                visitNode(yesNodes.item(i));
            }
            Node no = nodes.item(3);
            if (!NO.equals(no.getNodeName())) {
                throw new SchemeBuilderException("Irregular XML file. Excepted node "
                        + NO + ", but received " + no.getNodeName() + ".");
            }
            myState = BuildSchemeState.NO_BLOCK;
            myCurrentBlock = ib;
            NodeList noNodes = no.getChildNodes();
            n = noNodes.getLength();
            for (int i = 1; i < n; i += 2) {
                visitNode(noNodes.item(i));
            }
            myState = BuildSchemeState.BEGIN_STATE;
            myCurrentBlock = ib;
        } else if (INVOCATION_BLOCK.equals(blockType)) {
            IInvocationBlock ib = myBlockFactory.createInvocationBlock(blockText);
            addBlock(ib);
        } else if (SWITCH_BLOCK.equals(blockType)) {
            ISwitchBlock sb = myBlockFactory.createSwitchBlock(blockText);
            addBlock(sb);
            NodeList nodes = node.getChildNodes();
            int n = nodes.getLength();
            for (int i = 1; i < n; i += 2) {
                myCurrentBlock = sb;
                Node caseNode = nodes.item(i);
                if (!CASE.equals(caseNode.getNodeName())) {
                    throw new SchemeBuilderException("Irregular XML file. Excepted node "
                            + CASE + ", but received " + caseNode.getNodeName() + ".");
                }
                NamedNodeMap caseAttrs = caseNode.getAttributes();
                String caseText = caseAttrs.getNamedItem(TEXT).getNodeValue();
                sb.addCase(caseText);
                NodeList caseNodes = caseNode.getChildNodes();
                myState = BuildSchemeState.SWITCH_BLOCK;
                int m = caseNodes.getLength();
                for (int j = 1; j < m; j += 2) {
                    visitNode(caseNodes.item(j));
                }
            }
            myState = BuildSchemeState.BEGIN_STATE;
            myCurrentBlock = sb;
        }else if (TRY_BLOCK.equals(blockType)) {
            ITryBlock tb = myBlockFactory.createTryBlock();
            addBlock(tb);
            myState = BuildSchemeState.TRY_BLOCK;
            myCurrentBlock = tb;
            NodeList nodes = node.getChildNodes();
            int n = nodes.getLength();
            for (int i = 1; i < n; i += 2) {
                Node child = nodes.item(i);
                if (CATCH.equals(child.getNodeName())) {
                    NamedNodeMap catchAttrs = child.getAttributes();
                    String catchText = catchAttrs.getNamedItem(TEXT).getNodeValue();
                    IBlock cb = myBlockFactory.createActionBlock(catchText);
                    tb.addCatch(cb);
                    NodeList catchNodes = child.getChildNodes();
                    myState = BuildSchemeState.BEGIN_STATE;
                    myCurrentBlock = cb;
                    int m = catchNodes.getLength();
                    for (int j = 1; j < m; j += 2) {
                        visitNode(catchNodes.item(j));
                    }
                } else if (FINALLY.equals(child.getNodeName())) {
                    myState = BuildSchemeState.FINALLY;
                    myCurrentBlock = tb;
                    NodeList finallyNodes = child.getChildNodes();
                    int m = finallyNodes.getLength();
                    for (int j = 1; i < m; i += 2) {
                        visitNode(finallyNodes.item(j));
                    }
                } else {
                    visitNode(child);
                }
            }
            myState = BuildSchemeState.BEGIN_STATE;
            myCurrentBlock = tb;
        } else if (WHILE_BLOCK.equals(blockType)) {
            IWhileBlock wb = myBlockFactory.createWhileBlock(blockText);
            addBlock(wb);
            myState = BuildSchemeState.FIRST_WHILE_BLOCK;
            NodeList nodes = node.getChildNodes();
            int n = nodes.getLength();
            for (int i = 1; i < n; i += 2) {
                visitNode(nodes.item(i));
            }
            myState = BuildSchemeState.BEGIN_STATE;
            myCurrentBlock = wb;
        }
    }
    
    private void addBlock(IBlock block) {
        switch (myState) {
        case BEGIN_STATE:
        {
            myCurrentBlock.setNext(block);
            break;
        }
        case YES_BLOCK:
        {
            IIfBlock ib = (IIfBlock) myCurrentBlock;
            ib.setYesBlock(block);
            break;
        }
        case NO_BLOCK:
        {
            IIfBlock ib = (IIfBlock) myCurrentBlock;
            ib.setNoBlock(block);
            break;
        }
        case FIRST_DO_BLOCK:
        {
            IDoBlock dwb = (IDoBlock) myCurrentBlock;
            dwb.setFirstBlock(block);
            break;
        }
        case FIRST_FOR_BLOCK:
        {
            IForBlock fb = (IForBlock) myCurrentBlock;
            fb.setFirstBlock(block);
            break;
        }
        case FIRST_WHILE_BLOCK:
        {
            IWhileBlock wb = (IWhileBlock) myCurrentBlock;
            wb.setFirstBlock(block);
            break;
        }
        case SWITCH_BLOCK:
        {
            ISwitchBlock sb = (ISwitchBlock) myCurrentBlock;
            sb.setBlockForLastCase(block);
            break;
        }
        case TRY_BLOCK:
        {
            ITryBlock tb = (ITryBlock) myCurrentBlock;
            tb.setFirstBlock(block);
            break;
        }
        case FINALLY:
        {
            ITryBlock tb = (ITryBlock) myCurrentBlock;
            tb.setFinally(block);
            break;
        }
        }
        myCurrentBlock = block;
        myState = BuildSchemeState.BEGIN_STATE;
    }
    
}
