package org.amse.ak.schemebuilder.io.writer;

import org.amse.ak.schemebuilder.io.SchemeIOInterface;
import org.amse.ak.schemebuilder.model.IActionBlock;
import org.amse.ak.schemebuilder.model.IBeginBlock;
import org.amse.ak.schemebuilder.model.IBlock;
import org.amse.ak.schemebuilder.model.ICase;
import org.amse.ak.schemebuilder.model.IDoBlock;
import org.amse.ak.schemebuilder.model.IEndBlock;
import org.amse.ak.schemebuilder.model.IForBlock;
import org.amse.ak.schemebuilder.model.IIfBlock;
import org.amse.ak.schemebuilder.model.IInvocationBlock;
import org.amse.ak.schemebuilder.model.ISwitchBlock;
import org.amse.ak.schemebuilder.model.ITryBlock;
import org.amse.ak.schemebuilder.model.IWhileBlock;
import org.amse.ak.schemebuilder.model.ISchemeVisitor;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class WriteVisitor implements ISchemeVisitor, SchemeIOInterface {
    
    private final Document myDoc;
    
    private Element myCurrentElement;
    
    public WriteVisitor(Document doc, Element scheme) {
        myDoc = doc;
        myCurrentElement = scheme;
    }
    
    public void caseActionBlock(IActionBlock actionBlock) {
        if (actionBlock != null) {
            Element ae = myDoc.createElement(ACTION_BLOCK);
            ae.setAttribute(TEXT, actionBlock.getText());
            myCurrentElement.appendChild(ae);
            if (actionBlock.getNext() != null) {
                actionBlock.getNext().apply(this);
            }
        }
    }

    public void caseBeginBlock(IBeginBlock beginBlock) {
        if (beginBlock != null) {
            Element be = myDoc.createElement(BEGIN_BLOCK);
            be.setAttribute(TEXT, beginBlock.getText());
            myCurrentElement.appendChild(be);
            if (beginBlock.getNext() != null) {
                beginBlock.getNext().apply(this);
            }
        }
    }

    public void caseDoBlock(IDoBlock doBlock) {
        if (doBlock != null) {
            Element de = myDoc.createElement(DO_BLOCK);
            de.setAttribute(TEXT, doBlock.getText());
            myCurrentElement.appendChild(de);
            if (doBlock.getFirstBlock() != null) {
                Element outerElement = myCurrentElement;
                myCurrentElement = de;
                doBlock.getFirstBlock().apply(this);
                myCurrentElement = outerElement;
            }
            if (doBlock.getNext() != null) {
                doBlock.getNext().apply(this);
            }
        }
    }
    
    public void caseEndBlock(IEndBlock endBlock) {
        if (endBlock != null) {
            Element eb = myDoc.createElement(END_BLOCK);
            eb.setAttribute(TEXT, endBlock.getText());
            myCurrentElement.appendChild(eb);
        }
    }

    public void caseForBlock(IForBlock forBlock) {
        if (forBlock != null) {
            Element fe = myDoc.createElement(FOR_BLOCK);
            fe.setAttribute(TEXT, forBlock.getText());
            myCurrentElement.appendChild(fe);
            if (forBlock.getFirstBlock() != null) {
                Element outerElement = myCurrentElement;
                myCurrentElement = fe;
                forBlock.getFirstBlock().apply(this);
                myCurrentElement = outerElement;
            }
            if (forBlock.getNext() != null) {
                forBlock.getNext().apply(this);
            }
        }
    }

    public void caseIfBlock(IIfBlock ifBlock) {
        if (ifBlock != null) {
            Element ie = myDoc.createElement(IF_BLOCK);
            ie.setAttribute(TEXT, ifBlock.getText());
            myCurrentElement.appendChild(ie);
            Element yes = myDoc.createElement(YES);
            ie.appendChild(yes);
            if (ifBlock.getYesBlock() != null) {
                Element outerElement = myCurrentElement;
                myCurrentElement = yes;
                ifBlock.getYesBlock().apply(this);
                myCurrentElement = outerElement;
            }
            Element no = myDoc.createElement(NO);
            ie.appendChild(no);
            if (ifBlock.getNoBlock() != null) {
                Element outerElement = myCurrentElement;
                myCurrentElement = no;
                ifBlock.getNoBlock().apply(this);
                myCurrentElement = outerElement;
            }
            if (ifBlock.getNext() != null) {
                ifBlock.getNext().apply(this);
            }
        }
    }
    
    public void caseInvocationBlock(IInvocationBlock invocationBlock) {
        if (invocationBlock != null) {
            Element ib = myDoc.createElement(INVOCATION_BLOCK);
            ib.setAttribute(TEXT, invocationBlock.getText());
            myCurrentElement.appendChild(ib);
            if (invocationBlock.getNext() != null) {
                invocationBlock.getNext().apply(this);
            }
        }
    }

    public void caseSwitchBlock(ISwitchBlock switchBlock) {
        if (switchBlock != null) {
            Element se = myDoc.createElement(SWITCH_BLOCK);
            se.setAttribute(TEXT, switchBlock.getText());
            myCurrentElement.appendChild(se);
            for(ICase c: switchBlock.cases()) {
                Element ce = myDoc.createElement(CASE);
                ce.setAttribute(TEXT, c.getCase());
                se.appendChild(ce);
                if (c.getBlock() != null) {
                    Element outerElement = myCurrentElement;
                    myCurrentElement = ce;
                    c.getBlock().apply(this);
                    myCurrentElement = outerElement;
                }
            }
            if (switchBlock.getNext() != null) {
                switchBlock.getNext().apply(this);
            }
        }
    }
    
    public void caseTryBlock(ITryBlock tryBlock) {
        if (tryBlock != null) {
            Element te = myDoc.createElement(TRY_BLOCK);
            te.setAttribute(TEXT, tryBlock.getText());
            myCurrentElement.appendChild(te);
            if (tryBlock.getFirstBlock() != null) {
                Element outerElement = myCurrentElement;
                myCurrentElement = te;
                tryBlock.getFirstBlock().apply(this);
                myCurrentElement = outerElement;
            }
            for (IBlock catchBlock: tryBlock.getCatch()) {
                Element ce = myDoc.createElement(CATCH);
                ce.setAttribute(TEXT, catchBlock.getText());
                te.appendChild(ce);
                if (catchBlock.getNext() != null) {
                    Element outerElement = myCurrentElement;
                    myCurrentElement = ce;
                    catchBlock.getNext().apply(this);
                    myCurrentElement = outerElement;
                }
            }
            if (tryBlock.getFinally() != null) {
                Element fe = myDoc.createElement(FINALLY);
                te.appendChild(fe);
                Element outerElement = myCurrentElement;
                myCurrentElement = fe;
                tryBlock.getFinally().apply(this);
                myCurrentElement = outerElement;
            }
            if (tryBlock.getNext() != null) {
                tryBlock.getNext().apply(this);
            }
        }
    }
    
    public void caseWhileBlock(IWhileBlock whileBlock) {
        if (whileBlock != null) {
            Element we = myDoc.createElement(WHILE_BLOCK);
            we.setAttribute(TEXT, whileBlock.getText());
            myCurrentElement.appendChild(we);
            if (whileBlock.getFirstBlock() != null) {
                Element outerElement = myCurrentElement;
                myCurrentElement = we;
                whileBlock.getFirstBlock().apply(this);
                myCurrentElement = outerElement;
            }
            if (whileBlock.getNext() != null) {
                whileBlock.getNext().apply(this);
            }
        }
    }

}
